from discord_webhook import DiscordWebhook

def send(message: str, url):
    """Appelle le webhook Discord de l'url <url> avec le message <message>."""
    webhook = DiscordWebhook(url=url, content=message)
    response = webhook.execute()
    return response



import sys
import os

from dotenv import load_dotenv


# À l'import
if len(sys.argv) > 1 and sys.argv[1].startswith(".env"):
    path = sys.argv[1]
else:
    path = ".env"

load_dotenv(path)


def load(VAR_NAME):
    """Récupère la variable <VAR_NAME> depuis le .env demandé à l'appel de bot.py (.env par défaut)"""
    VAR = os.getenv(VAR_NAME)
    assert VAR, f"Variable d'environnement {VAR_NAME} manquante"
    return VAR



import sys
import re
import traceback
import asyncio
import keyword


### AVERTISSMENT : le code ci-dessous est peu commenté et pas mal technique (full orienté objet et quelques regex)


class PseudoShellExit(BaseException):
    """Ordonne la fermeture du pseudo-terminal Python"""

    def __repr__(self):
        """Return repr(self)"""
        return f"<pseudoshell.PseudoShellExit {BaseException.__repr__(self)}>"


class ShellIOError(PseudoShellExit):
    """Erreur lors d'une entrée / sortie du pseudo-terminal Python"""

    def __init__(self, method, original):
        """Initialize self"""
        PseudoShellExit.__init__(self, f"PseudoShell : Exception in {method}, exited.")
        self.method = method            # Méthode ayant levé l'exception
        self.original = original        # Exception originale

    def __repr__(self):
        """Return repr(self)"""
        return f"<pseudoshell.ShellIOError: Exception raised on Shell.{self.method}: \"{self.original}\">"


class Shell(object):          # Les attributs de cette classe peuvent être modifiés via exec
    """Pseudo-terminal Python"""

    def __init__(self, globals, locals, pull=input, push=print, welcome_text="", shut_keywords=[],
                 bridge_name="_shell", result_name="_", coro_name="_coro", prov_name="_prov", var_name="_var"):
        """Initialize self"""
        self.globals = globals                  # Variables globales de l'environnement d'exécution (dictionnaire)
        self.locals = locals                    # Variables locales de l'environnement d'exécution (dictionnaire)
        self._pull = pull                       # Fonction / coroutine None -> str d'input
        self._push = push                       # Fonction / coroutine str -> None d'output
        self.welcome_text = welcome_text        # Texte affiché au lancement
        self.shut_keywords = shut_keywords      # Liste des mots-clés réservés (str), coupant l'exécution sans appeller le terminal
        self.bridge_name = bridge_name          # Nom de la variable associée à self dans les locals des exec
        self.result_name = result_name          # Nom de la variable utilisée pour récupérer les résultats des exec
        self.coro_name = coro_name              # Nom de la variable utilisée pour la gestion des coroutines des exec
        self.prov_name = prov_name              # Nom de la variable utilisée pour diverses utilités temporaires
        self.var_name = var_name                # Nom de la variable utilisée pour des affectations extérieures

        self.locals[self.bridge_name] = self    # Permet d'accéder au shell dans les exec

    async def pull(self):
        """Retourne une entrée utilisateur (normalement str) dans le pseudo-terminal"""
        try:
            result = self._pull()
            if asyncio.iscoroutine(result):
                result = await result
        except Exception as exc:
            raise ShellIOError(method="pull", original=exc)
        else:
            return result

    async def push(self, text, color=False):
        """Envoie <text> dans le pseudo-terminal"""
        try:
            result = self._push(text, color)
            if asyncio.iscoroutine(result):
                result = await result
        except Exception as exc:
            raise ShellIOError(method="push", original=exc)

    async def help(self):
        """Envoie le message d'aide d'utilisation du pseudo-terminal"""
        await self.push(
            """Bon flemme d'écrire toute la doc' pour ça, en gros il faut bien comprendre que seules les syntaxes simples marcheront :
    - assignation           (var = expression)
    - évaluation            (expression)
    - instruction for       fin avec "end" ou "endfor", clause "else" reconnue
    - instruction while     fin avec "end" ou "endwhile" clause "else" reconnue
    - instruction if        fin avec "end" ou "endif", clauses "elif" et "else" reconnues

Les insctuctions try / (async) with / (async) def / class / async for ne sont pas disponibles pour l'instant, notemment.
Les clauses break et continue ne sont pas implémentées et soulèvent une erreur d'exécution.

Chaque ligne est exécutée séparément : il est impossible de séparer des instructions en plusieurs lignes (\, \"\"\", continuations de parenthèses...)"""
        )

    async def run(self):
        """Lance le pseudo-terminal"""
        # Pré-vérifications
        if "sys" not in self.globals:
            self.globals["sys"] = sys       # On doit avoir sys importé dans shell.globals

        # Run
        version_info = sys.version.replace('\n', '')
        if self.shut_keywords:
            shut_keywords = ", ".join(f"\"{kw}\"" for kw in self.shut_keywords) + " (arrêt immédiat), "
        else:
            shut_keywords = ""
        await self.push(f"""Python {version_info}\n{self.welcome_text}\n"""
                        f"""Mots-clés réservés : {shut_keywords}"end", "endfor", "endif", "endwhile" (sortie de structure), \"{self.bridge_name}", \"{self.result_name}", \"{self.coro_name}", \"{self.prov_name}" et \"{self.var_name}".\n"""
                        f"""Utiliser "help" pour plus d'informations sur les spécificités de ce pseudo-terminal (help(obj) non modifié)"""
        )

        main_loop = Loop(shell=self, depth=0)
        await main_loop.run()

        await self.push(f"""Exiting pseudo-shell, bye""")


class Result():
    """Résultat d'une exécution dans un pseudo-terminal"""

    def __init__(self, arg, success=None):
        """Initialize self"""
        if isinstance(arg, str):
            self.text = arg
            self.success = success or True                  # par défaut : succès
        else:
            raise TypeError(f"TypeError: Result() argument must be a string instance, not '{type(arg)}'")

    def __repr__(self):
        """Return repr(self)"""
        return f"""<pseudoshell.Result \"{self.text}" (success={self.success})>"""

    def __bool__(self):
        """Return self.text.__bool__"""
        return bool(self.text)


class ResultsList(list):
    """Liste d'objets Result"""
    def __init__(self, iterable=(), /, success=None):
        """Initialize self"""
        try:
            if all(isinstance(obj, Result) for obj in iterable):    # Si que des Result:
                list.__init__(self, iterable)
                if success is None:        # Si la liste n'est pas vide
                    self.success = self[-1].success if self else True
                else:
                    self.success = success
            else:
                raise TypeError
        except (TypeError, ValueError) as e:
            raise TypeError("TypeError: ResultsList() argument must be an iterator of Result instances") from e

    def __repr__(self):
        """Return repr(self)"""
        return f"""<pseudoshell.ResultsList {list.__repr__(self)}>"""

    def append(self, item):
        """Implémente list.append(item) ; self.success = item.success"""
        if isinstance(item, Result):
            list.append(self, item)
            self.success = item.success
        else:
            raise TypeError("TypeError: ResultsList.append() argument must be a Result instance")

    def extend(self, iterable):
        """Implémente list.extend(results_list) ; self.success = iterable.success"""
        try:
            if not isinstance(iterable, ResultsList):
                try:
                    iterable = ResultsList(iterable)
                except:
                    raise
            list.extend(self, iterable)
            self.success = iterable.success
        except (TypeError, ValueError) as e:
            raise TypeError("TypeError: ResultsList.extend() argument must be a ResultsList instance or any iterator of Result instances") from e

    def join(self, sep="\n") -> Result:
        """Joins les résultats dans la liste -> Result"""
        text = sep.join(result.text for result in self)
        return Result(text, success=self.success)


class PseudoShellError(Exception):
    """Erreur du pseudo-terminal Python"""

    def __repr__(self):
        """Return repr(self)"""
        return f"<pseudoshell.PseudoShellError {Exception.__repr__(self)}>"


class ExecutionError(PseudoShellError):
    """Erreur à l'exécution d'une ligne de pseudo-terminal Python"""

    def __init__(self, original):
        """Initialize self"""
        Exception.__init__(self)
        self.original = original        # Exception originale

    def __repr__(self):
        """Return repr(self)"""
        return f"<pseudoshell.ExecutionError original=\"{self.original}\">"

    def original_traceback(self):
        """Retourne le traceback de l'exception levée pendant l'exécution"""
        return "".join(s for i,s in enumerate(traceback.format_exception(type(self.original), self.original, self.original.__traceback__)) if i != 1)       # On extrait la ligne 1, qui contient les infos avant d'entrer dans le shell (appel d'exec())


class Loop():
    """Boucle de pseudo-terminal Python"""

    def __init__(self, shell, depth, caller="", history=""):
        """Initialize self"""
        self.shell = shell
        self.depth = depth
        self.caller = caller          # Ligne d'appel de la boucle
        self.history = history        # Lignes / structures pas encore exécutées
        self.prompt = ">>> " + "    "*self.depth
        self.prompt_dots = ">>> " + "··· "*self.depth + "        << : end"
        self.self_prompt = ">>> " + "    "*(self.depth - 1)
        self.endwords = ["end"]
        self.buffer = []               # Lignes en attende d'exécution
        self.exit = False

    def __repr__(self):
        """Return repr(self)"""
        return f"""<pseudoshell.{type(self).__name__} \"{self.caller}" on shell {self.shell}>"""

    async def run(self):
        """Exécute la boucle jusqu'à recevoir un ordre de fin"""
        if self.history:
            await self.shell.push(self.history, color=True)

        while not self.exit:
            await self.shell.push(self.prompt_dots, color=True)
            entree = await self.shell.pull()

            try:
                if entree in self.endwords:
                    self.history += f"{self.self_prompt}\n"
                    self.exit = True

                elif entree == "help":
                    await self.shell.help()

                elif entree.endswith(":"):      # Commande d'entrée structure / clause

                    # Structures
                    new_struct = None
                    if entree.startswith("for"):
                        new_struct = For
                    elif entree.startswith("while"):
                        new_struct = While
                    elif entree.startswith("if"):
                        new_struct = If

                    if new_struct:
                        loop = new_struct(shell=self.shell, caller=entree, depth=self.depth + 1,
                                          history=f"{self.history}{self.prompt}{entree}\n")
                        await loop.run()                    # Récupération instructions
                        self.buffer.append(loop)            # Une fois boucle finie, on l'enregistre (objet If)
                        self.history = loop.history         # On actualise l'historique


                    # Clauses des différentes structures
                    elif entree.startswith("else"):
                        self.add_clause("else", entree)
                        self.history += f"{self.self_prompt}{entree}\n"
                    elif entree.startswith("elif"):
                        self.add_clause("elif", entree)
                        self.history += f"{self.self_prompt}{entree}\n"

                    else:
                        raise PseudoShellError(f"BadSyntax - Clause non reconnue")


                else:       # Si ni for, while, if, ..., exit :
                    self.buffer.append(Line(shell=self.shell, text=entree))
                    self.history += f"{self.prompt}{entree}\n"


                if self.depth == 0:       # Boucle principale : on exécute
                    try:
                        result = await self.exec()

                    except ExecutionError as exc:
                        await self.shell.push(f"{exc.original_traceback()}")

                    except Exception:
                        await self.shell.push(f"PseudoShell - Fatal exception during execution :\n{traceback.format_exc()}\n"
                                              f"No change made, retry:")
                        await self.shell.push(self.history, color=True)

                    else:
                        await self.shell.push(self.history, color=True)
                        await self.shell.push(result.text)
                        # await self.shell.push(f"{self.history}{result.text}\n\nbuffer : {self.buffer}")

                    finally:
                        self.buffer = []
                        self.history = ""

                else:                     # Dans une structure : on enregistre
                    await self.shell.push(self.history, color=True)

            except Exception as exc:
                await self.shell.push(f"Trying to add\n{self.prompt}{entree}\n"
                                      f"{''.join(traceback.format_exception_only(type(exc), exc))}\n"
                                      f"No change made, retry:"
                )
                await self.shell.push(self.history, color=True)


    def add_clause(self, clause, caller):
        """Gère l'ajout de la clause <clause> à la structure. À overwrite."""
        raise PseudoShellError(f"BadSyntax - Clause {clause} incorrecte / non appliquable ici")

    async def exec_buffer(self, start=None, end=None) -> Result:
        """Exécute les objets et structures dans self.buffer[start:end] (défaut tous)"""
        results = ResultsList()
        for obj in self.buffer[start:end]:   # lignes et structures dans buffer
            result = await obj.exec()               # ON EXÉCUTE TOUT
            if result:
                results.append(result)

        return results.join()      # Joins les results entre eux (garde le success du dernier result)

    exec = exec_buffer      # Pour la boucle "de base", ça revient au même


class For(Loop):
    """Boucle for d'un pseudo-terminal Shell"""

    def __init__(self, *args, **kwargs):
        """Initialize self"""
        Loop.__init__(self, *args, **kwargs)
        self.endwords.append("endfor")
        self.prompt_dots += ", endfor"

        if match := re.match(r"for\s*(.+?)\s*in\s*(.+?)\s*:", self.caller):     # Entrée dans boucle
            self.variter_txt = match.group(1).strip()
            self.iterator_txt = match.group(2).strip()
        else:
            raise PseudoShellError("BadSyntax - Mauvaise syntaxe pour une boucle for")

        self.else_index = None


    def add_clause(self, clause, caller):
        """Enregistre la clause <clause>"""
        if clause == "else":
            if self.else_index:
                raise PseudoShellError("BadSyntax - Il ne peux y avoir plus d'une clause else")
            elif match := re.match(r"else\s*:", caller):
                self.else_index = len(self.buffer)          # Indice de début de clause else
            else:
                raise PseudoShellError("BadSyntax - Mauvaise syntaxe pour une clause else")

        else:
            raise PseudoShellError(f"BadSyntax - Clause {clause} incorrecte dans if")

    async def exec(self) -> Result:
        """Exécute la boucle"""
        text = f"{self.shell.prov_name} = {self.iterator_txt}"
        await Line(shell=self.shell, text=text).exec()      # On affecte l'itérateur à _prov (dans le contexte de shell)

        text = f"{self.variter_txt} = {self.shell.var_name}"
        results = ResultsList()
        for self.shell.locals[self.shell.var_name] in self.shell.locals[self.shell.prov_name]:   # for _var in _prov
            await Line(shell=self.shell, text=text).exec()                  # On affecte _var à {variter} (dans le contexte de shell)
            iter_result = await self.exec_buffer(end=self.else_index)       # Résultat de l'itération en cours
            results.append(iter_result)

        if self.else_index:
            iter_result = await self.exec_buffer(start=self.else_index)      # Exécution clause else
            results.append(iter_result)

        return results.join()      # Joins les results entre eux (crée un Result, garde le success du dernier result de la liste)


class While(Loop):
    """Boucle while d'un pseudo-terminal Shell"""

    def __init__(self, *args, **kwargs):
        """Initialize self"""
        Loop.__init__(self, *args, **kwargs)
        self.endwords.append("endwhile")
        self.prompt_dots += ", endwhile"

        if match := re.match(r"while\s*(.+?)\s*:", self.caller):     # Entrée dans boucle
            self.condition = match.group(1).strip()
        else:
            raise PseudoShellError("BadSyntax - Mauvaise syntaxe pour une boucle while")

        self.else_index = None

    def add_clause(self, clause, caller):
        """Enregistre la clause <clause>"""
        if clause == "else":
            if self.else_index:
                raise PseudoShellError("BadSyntax - Il ne peux y avoir plus d'une clause else")
            elif match := re.match(r"else\s*:", caller):
                self.else_index = len(self.buffer)          # Indice de début de clause else
            else:
                raise PseudoShellError("BadSyntax - Mauvaise syntaxe pour une clause else")

        else:
            raise PseudoShellError(f"BadSyntax - Clause {clause} incorrecte dans while")

    async def exec(self) -> Result:
        """Exécute la boucle"""
        text = f"{self.shell.prov_name} = {self.condition}"
        await Line(shell=self.shell, text=text).exec()      # On affecte la condition à _prov (dans le contexte de shell)

        results = ResultsList()
        while self.shell.locals[self.shell.prov_name]:
            iter_result = await self.exec_buffer(end=self.else_index)       # Résultat de l'itération en cours
            results.append(iter_result)

            await Line(shell=self.shell, text=text).exec()      # On recalcule la condition

        if self.else_index:
            iter_result = await self.exec_buffer(start=self.else_index)      # Exécution clause else
            results.append(iter_result)

        return results.join()      # Joins les results entre eux (crée un Result, garde le success du dernier result de la liste)



class If(Loop):
    """Boucle if (niklérajeu) d'un pseudo-terminal Shell"""

    def __init__(self, *args, **kwargs):
        """Initialize self"""
        Loop.__init__(self, *args, **kwargs)
        self.endwords.append("endif")
        self.prompt_dots += ", endif"

        if match := re.match(r"if\s*(.+?)\s*:", self.caller):
            self.conditions = [match.group(1).strip()]
            self.indexes = [0]
        else:
            raise PseudoShellError("BadSyntax - Mauvaise syntaxe pour une clause if")

        self.else_index = None


    def add_clause(self, clause, caller):
        """Enregistre la clause <clause>"""
        if clause == "else":
            if self.else_index:
                raise PseudoShellError("BadSyntax - Il ne peux y avoir plus d'une clause else")
            elif match := re.match(r"else\s*:", caller):
                self.else_index = len(self.buffer)          # Indice de début de clause else
            else:
                raise PseudoShellError("BadSyntax - Mauvaise syntaxe pour une clause else")

        elif clause == "elif":
            if self.else_index:
                raise PseudoShellError("BadSyntax - Une clause elif ne peut venir après une clause else")
            elif match := re.match(r"elif\s*(.+?)\s*:", caller):
                self.conditions.append(match.group(1).strip())
                self.indexes.append(len(self.buffer))       # Indice de début de le clause elif
            else:
                raise PseudoShellError("BadSyntax - Mauvaise syntaxe pour une clause elif")

        else:
            raise PseudoShellError(f"BadSyntax - Clause {clause} incorrecte dans if")


    async def exec(self) -> Result:
        """Exécute la boucle"""
        self.indexes.append(self.else_index)       # P'tit trix
        for i, condition in enumerate(self.conditions):     # Pour chaques elifs (y compris le if initial)
            text = f"{self.shell.prov_name} = {condition}"
            await Line(shell=self.shell, text=text).exec()      # On affecte la condition à _prov (dans le contexte de shell)

            if self.shell.locals[self.shell.prov_name]:
                results = await self.exec_buffer(start=self.indexes[i], end=self.indexes[i+1])
                return results

        if self.else_index:
            results = await self.exec_buffer(start=self.else_index)
            return results



class Line():
    """Ligne de commande Python reçue de l'utilisateur"""

    def __init__(self, shell, text):
        """Initialize self"""
        self.text = text
        self.shell = shell
        self.result = Result("")

    async def exec(self) -> Result:
        """Évalue la ligne dans le contexte de self.shell (globals et locals)"""
        self.result = Result("")        # On réinitialise pour pouvoir exécuter la ligne plusieurs fois
        stdout = sys.stdout
        self.shell.globals["sys"].stdout = self      # Petit hack pour envoyer print() vers self.result (via self.write)

        source = self.text

        # Gestion coroutine
        if "await" in source:
            left, right = source.split("await", maxsplit=1)                             # Ex. left = "mess = ", right = " ctx.send('oh')"
            res = await Line(self.shell, f"{self.shell.coro_name} = {right}").exec()    # On exec la partie droite  ==> _coro = <coroutine object ...>
            self.shell.locals[self.shell.coro_name] = await self.shell.locals[self.shell.coro_name]     #   On await ==> _coro = <Message id=...>
            source = f"{left} {self.shell.coro_name}"                                   #  Notre nouvelle source, ex. : "mess =  _coro"

        # Préparation récupération résultat
        if not ("=" in source
                or keyword.iskeyword(source.split()[0])):       # Ligne à exécuter
            source = f"{self.shell.result_name} = {source}"

        # Exécution
        source = f"{source} ; {self.shell.bridge_name}.globals = globals() ; {self.shell.bridge_name}.locals = locals()"
        try:
            exec(source, self.shell.globals, self.shell.locals)

        except Exception as exc:
            raise ExecutionError(original=exc)

        except SystemExit:
            raise PseudoShellExit("PseudoShell : exited.")

        else:
            # Récupération résultat
            if self.shell.result_name in self.shell.locals:
                if not self.result:
                    self.result.text = repr(self.shell.locals[self.shell.result_name])
                self.shell.locals.pop(self.shell.result_name)

            return self.result

        finally:
            self.shell.globals["sys"].stdout = stdout


    def write(self, text):
        """Petit hack pour pouvoir rediriger print vers self, élimine les \n envoyés seuls à la fin"""
        if text != "\n":
            self.result.text += text



import traceback

from discord.ext import commands

from lgrez.blocs import tools, bdd, bdd_tools, env, gsheets
from lgrez.blocs.bdd import Joueurs, Actions, BaseActions, BaseActionsRoles, Taches
from lgrez.features import gestion_actions


async def get_sync():
    """Récupère les modifications en attente sur le TDB"""

    cols = [col for col in bdd_tools.get_cols(Joueurs) if not col.startswith('_')]    # On élimine les colonnes locales
    cols_SQL_types = bdd_tools.get_SQL_types(Joueurs)
    cols_SQL_nullable = bdd_tools.get_SQL_nullable(Joueurs)

    ### RÉCUPÉRATION INFOS GSHEET

    SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")

    workbook = gsheets.connect(SHEET_ID)    # Tableau de bord
    sheet = workbook.worksheet("Journée en cours")
    values = sheet.get_all_values()         # Liste de liste des valeurs des cellules
    (NL, NC) = (len(values), len(values[0]))

    head = values[2]            # Ligne d'en-têtes (noms des colonnes) = 3e ligne du TDB
    TDB_index = {col: head.index(col) for col in cols}    # Dictionnaire des indices des colonnes GSheet pour chaque colonne de la table
    TDB_tampon_index = {col: head.index(f"tampon_{col}") for col in cols if col != 'discord_id'}    # Idem pour la partie « tampon »

    # CONVERSION INFOS GSHEET EN UTILISATEURS

    joueurs_TDB = []            # Liste des joueurs tels qu'actuellement dans le TDB
    ids_TDB = []                # discord_ids des différents joueurs du TDB
    rows_TDB = {}               # Indices des lignes ou sont les différents joueurs du TDB

    for l in range(NL):
        L = values[l]           # On parcourt les lignes du TDB
        id_cell = L[TDB_index["discord_id"]]
        if id_cell.isdigit():        # Si la cellule contient bien un ID (que des chiffres, et pas vide)
            id = int(id_cell)
            joueur_TDB = {col: bdd_tools.transtype(L[TDB_index[col]], col, cols_SQL_types[col], cols_SQL_nullable[col]) for col in cols}
                # Dictionnaire correspondant à l'utilisateur
            joueurs_TDB.append(joueur_TDB)
            ids_TDB.append(id)
            rows_TDB[id] = l

    ### RÉCUPÉRATION UTILISATEURS CACHE

    joueurs_BDD = Joueurs.query.all()     # Liste des joueurs tels qu'actuellement en cache
    ids_BDD = [joueur_BDD.discord_id for joueur_BDD in joueurs_BDD]

    ### COMPARAISON

    modifs = []         # modifs à porter au TDB : tuple (id - colonne (nom) - valeur)
    modified_ids = []

    for joueur_BDD in joueurs_BDD.copy():                   ## Joueurs dans le cache supprimés du TDB
        if joueur_BDD.discord_id not in ids_TDB:
            joueurs_BDD.remove(joueur_BDD)
            bdd.session.delete(joueur_BDD)

    for joueur_TDB in joueurs_TDB:                              ## Différences
        id = joueur_TDB["discord_id"]

        if id not in ids_BDD:             # Si joueur dans le cache pas dans le TDB
            raise ValueError(f"Joueur {joueur_TDB['nom']} hors BDD : vérifier processus d'inscription")

        joueur_BDD = [joueur for joueur in joueurs_BDD if joueur.discord_id == id][0]     # joueur correspondant dans le cache

        for col in cols:
            if getattr(joueur_BDD, col) != joueur_TDB[col]:   # Si <col> diffère entre TDB et cache

                modifs.append( (id, col, joueur_TDB[col]) )   # On ajoute les modifs
                if id not in modified_ids:
                    modified_ids.append(id)

    ### APPLICATION DES MODIFICATIONS SUR LE TDB

    if modifs:
        modifs_lc = [(rows_TDB[id], TDB_tampon_index[col], v) for (id, col, v) in modifs]
            # On transforme les infos en coordonnées dans le TDB : ID -> ligne et col -> colonne,
        gsheets.update(sheet, modifs_lc)


    ### RETOURNAGE DES RÉSULTATS

    return {id: {col: v for (idM, col, v) in modifs if idM == id} for id in modified_ids}



async def modif_joueur(ctx, joueur_id, modifs, silent):
    """Attribue les modifications demandées au joueur

    modifs : dictionnaire {colonne BDD: nouvelle valeur}
    """
    joueur = Joueurs.query.get(int(joueur_id))
    assert joueur, f"!sync : joueur d'ID {joueur_id} introuvable"

    member = ctx.guild.get_member(joueur.discord_id)
    assert member, f"!sync : member {joueur} introuvable"

    chan = ctx.guild.get_channel(joueur._chan_id)
    assert chan, f"!sync : chan privé de {member} introuvable"

    changelog = f"\n- {member.display_name} (@{member.name}#{member.discriminator}) :\n"
    notif = ""

    for col, val in modifs.items():
        changelog += f"    - {col} : {val}\n"

        if col == "nom":                            # Renommage joueur
            await chan.edit(name=f"conv-bot-{val}")
            await member.edit(nick=val)
            if not silent:
                notif += f":arrow_forward: Tu t'appelles maintenant {tools.bold(val)}.\n"

        elif col == "chambre" and not silent:       # Modification chambre
            notif += f":arrow_forward: Tu habites maintenant en chambre {tools.bold(val)}.\n"

        elif col == "statut":
            if val == "vivant":                     # Statut = vivant
                await member.add_roles(tools.role(ctx, "Joueur en vie"))
                await member.remove_roles(tools.role(ctx, "Joueur mort"))
                if not silent:
                    notif += f":arrow_forward: Tu es maintenant en vie. EN VIE !!!\n"

            elif val == "mort":                     # Statut = mort
                await member.add_roles(tools.role(ctx, "Joueur mort"))
                await member.remove_roles(tools.role(ctx, "Joueur en vie"))
                if not silent:
                    notif += f":arrow_forward: Tu es malheureusement décédé(e) :cry:\nÇa arrive même aux meilleurs, en espérant que ta mort ait été belle !\n"
                # Actions à la mort
                for action in Joueurs.query.filter_by(player_id=joueur.discord_id, trigger_debut="mort"):
                    await gestion_actions.open_action(ctx, action, chan)

            elif val == "MV":                       # Statut = MV
                await member.add_roles(tools.role(ctx, "Joueur en vie"))
                await member.remove_roles(tools.role(ctx, "Joueur mort"))
                if not silent:
                    notif += f":arrow_forward: Te voilà maintenant réduit(e) au statut de mort-vivant... Un MJ viendra te voir très vite, si ce n'est déjà fait, mais retient que la partie n'est pas finie pour toi !\n"

            elif not silent:                        # Statut = autre
                notif += f":arrow_forward: Nouveau statut : {tools.bold(val)} !\n"

        elif col == "role":                         # Modification rôle
            old_bars = Joueurs.query.filter_by(role=joueur.role).all()
            old_actions = []
            for bar in old_bars:
                old_actions.extend(Joueurs.query.filter_by(action=bar.action, player_id=joueur.discord_id).all())
            for action in old_actions:
                gestion_actions.delete_action(ctx, action)  # On supprime les anciennes actions de rôle (et les tâches si il y en a)

            new_bars = Joueurs.query.filter_by(role=val).all()         # Actions associées au nouveau rôle
            new_bas = [Joueurs.query.get(bar.action) for bar in new_bars]   # Nouvelles BaseActions
            cols = [col for col in bdd_tools.get_cols(BaseActions) if not col.startswith("base")]
            new_actions = [Actions(player_id=joueur.discord_id, **{col: getattr(ba, col) for col in cols},
                                   cooldown=0, charges=ba.base_charges) for ba in new_bas]
            await tools.log(ctx, str(new_actions))

            for action in new_actions:
                gestion_actions.add_action(ctx, action)     # Ajout et création des tâches si trigger temporel

            role = tools.nom_role(val)
            if not role:        # role <val> pas en base : Error!
                role = f"« {val} »"
                await tools.log(ctx, f"{tools.mention_MJ(ctx)} ALED : rôle \"{val}\" attribué à {joueur.nom} inconnu en base !")
            if not silent:
                notif += f":arrow_forward: Ton nouveau rôle, si tu l'acceptes : {tools.bold(role)} !\nQue ce soit pour un jour ou pour le reste de la partie, renseigne toi en tapant {tools.code(f'!roles {val}')}.\n"

        elif col == "camp" and not silent:          # Modification camp
            notif += f":arrow_forward: Tu fais maintenant partie du camp « {tools.bold(val)} ».\n"

        elif col == "votant_village" and not silent:
            if val:                                 # votant_village = True
                notif += f":arrow_forward: Tu peux maintenant participer aux votes du village !\n"
            else:                                   # votant_village = False
                notif += f":arrow_forward: Tu ne peux maintenant plus participer aux votes du village.\n"

        elif col == "votant_loups" and not silent:
            if val:                                 # votant_loups = True
                notif += f":arrow_forward: Tu peux maintenant participer aux votes des loups ! Amuse-toi bien :wolf:\n"
            else:                                   # votant_loups = False
                notif += f":arrow_forward: Tu ne peux maintenant plus participer aux votes des loups.\n"

        elif col == "role_actif" and not silent:
            if val:                                 # role_actif = True
                notif += f":arrow_forward: Tu peux maintenant utiliser tes pouvoirs !\n"
            else:                                   # role_actif = False
                notif += f":arrow_forward: Tu ne peux maintenant plus utiliser aucun pouvoir.\n"

        bdd_tools.modif(joueur, col, val)           # Dans tous les cas, on modifie en base (après, pour pouvoir accéder aux vieux attribus plus haut)

    if not silent:
        await chan.send(":zap: Une action divine vient de modifier ton existence ! :zap:\n"
                        + f"\n{notif}\n"
                        + tools.ital(":warning: Si tu penses qu'il y a erreur, appelle un MJ au plus vite !"))

    return changelog


class Sync(commands.Cog):
    """Sync - Commandes de synchronisation des GSheets vers la BDD et les joueurs"""

    @commands.command()
    @tools.mjs_only
    async def sync(self, ctx, silent=False):
        """Applique les modifications lors d'un appel du Tableau de bord (COMMANDE MJ)

        <silent> peut valoir
            False / 0       les joueurs sont notifiés si leur statut est modifié (défaut)
            True  / 1       les joueurs ne sont pas notifiés

        Cette commande va récupérer les modifications en attente sur le Tableau de bord (lignes en rouge), modifer la BDD Joueurs, et appliquer les modificatons dans Discord le cas échéant : renommage des utilisateurs, modification des rôles...
        """
        try:
            await ctx.send("Récupération des modifications...")
            async with ctx.typing():
                dic = await get_sync()                  # Récupération du dictionnaire {joueur_id: modified_attrs}
                silent = (silent == "True")
                changelog = f"Synchronisation TDB (silencieux = {silent}) :"

            if dic:
                nb_modifs = sum(len(modifs) for modifs in dic.values())
                await ctx.send(f"{nb_modifs} modification(s) trouvée(s) pour {len(dic)} joueur(s), application...")
                async with ctx.typing():
                    for joueur_id, modifs in dic.items():               # Joueurs dont au moins un attribut a été modifié
                        try:
                            changelog += await modif_joueur(ctx, joueur_id, modifs, silent)
                        except Exception as e:
                            changelog += traceback.format_exc()
                            await ctx.send(f"Erreur joueur {joueur_id}, passage au suivant, voir logs pour les détails")

                    bdd.session.commit()

                    await tools.log(ctx, changelog, code=True)

                await ctx.send(f"Fait (voir {tools.channel(ctx, 'logs')} pour le détail)")

            else:
                await ctx.send("Pas de nouvelles modificatons.")

        except Exception:
            await tools.log(ctx, traceback.format_exc(), code=True)



import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from lgrez.blocs import env, bdd



Base = declarative_base()
Tables = {}


# Objets de connection (créés dans connect)
engine = None
session = None


class ClassProperty(property):      # https://stackoverflow.com/a/1383402
    """Permet d'utiliser des propriés comme des classmethods (reçoivent la classe comme seul argument)"""
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class MyTable(object):
    def __init_subclass__(cls, **kwargs):
        """Définit __tablename__ et enregistre la table dans Tables"""
        super().__init_subclass__(**kwargs)
        cls.__tablename__ = cls.__name__.lower()
        Tables[cls.__name__] = cls

    @classmethod
    def get_query(cls):
        return bdd.session.query(cls)

    query = ClassProperty(get_query)        # Permet d'utiliser Table.query.<...> au lieu de bdd.session.query(Table).<...>



# Définition des tables

class Joueurs(MyTable, Base): #, tablename="joueurs"
    """Table de données des joueurs inscrits"""

    discord_id = sqlalchemy.Column(sqlalchemy.BigInteger(), primary_key=True)
    _chan_id = sqlalchemy.Column(sqlalchemy.BigInteger(), nullable=False)

    nom = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)
    chambre = sqlalchemy.Column(sqlalchemy.String(200), nullable=False)
    statut = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    role = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)
    camp = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    votant_village = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=False)
    votant_loups = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=False)
    role_actif = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    _vote_condamne = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)
    _vote_maire = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)
    _vote_loups = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)

    def __repr__(self):
        """Return repr(self)."""
        return f"<Joueurs #{self.discord_id} ({self.nom})>"



class Roles(MyTable, Base): #, tablename="roles"
    """Table de données des rôles"""

    slug = sqlalchemy.Column(sqlalchemy.String(32), primary_key=True)

    prefixe = sqlalchemy.Column(sqlalchemy.String(8), nullable=False)
    nom = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    camp = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)                       # loups, solitaire, nécro, village...

    description_courte = sqlalchemy.Column(sqlalchemy.String(140), nullable=False)
    description_longue = sqlalchemy.Column(sqlalchemy.String(2000), nullable=False)

    def __repr__(self):
        """Return repr(self)."""
        return f"<Roles '{self.slug}' ({self.prefixe}{self.nom})>"



class BaseActions(MyTable, Base): #, tablename="base_actions"
    """Table de données des actions définies de bases (non liées à un joueur)"""

    action = sqlalchemy.Column(sqlalchemy.String(32), primary_key=True)

    trigger_debut = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    trigger_fin = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    instant = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    heure_debut = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)
    heure_fin = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)

    base_cooldown = sqlalchemy.Column(sqlalchemy.Integer(), nullable=False)
    base_charges = sqlalchemy.Column(sqlalchemy.Integer(), nullable=True)
    refill  = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)

    lieu = sqlalchemy.Column(sqlalchemy.String(32), nullable=True) #Distance/Physique/Lieu/Contact/Conditionnel/None/Public
    interaction_notaire = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)         # Oui, Non, Conditionnel, Potion, Rapport; None si récursif
    interaction_gardien = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)         # Oui, Non, Conditionnel, Taverne, Feu, MaisonClose, Précis, Cimetière, Loups, None si recursif
    mage = sqlalchemy.Column(sqlalchemy.String(100), nullable=True)                       #Oui, Non, changement de cible, etc

    changement_cible = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)


    def __repr__(self):
        """Return repr(self)."""
        return f"<BaseActions '{self.action}'>"



class Actions(MyTable, Base): #, tablename="actions"
    """Table de données des actions attribuées (liées à un joueur et actives)"""

    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    player_id = sqlalchemy.Column(sqlalchemy.BigInteger(), nullable=False)
    action = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    trigger_debut = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    trigger_fin = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    instant = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    heure_debut = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)
    heure_fin = sqlalchemy.Column(sqlalchemy.Time(), nullable=True)

    cooldown = sqlalchemy.Column(sqlalchemy.Integer(), nullable=False)
    charges = sqlalchemy.Column(sqlalchemy.Integer(), nullable=True)
    refill  = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)

    lieu = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    interaction_notaire = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    interaction_gardien = sqlalchemy.Column(sqlalchemy.String(32), nullable=True)
    mage = sqlalchemy.Column(sqlalchemy.String(100), nullable=True)
    changement_cible = sqlalchemy.Column(sqlalchemy.Boolean(), nullable=True)

    _decision = sqlalchemy.Column(sqlalchemy.String(200), nullable=True)


    def __repr__(self):
        """Return repr(self)."""
        return f"<Actions #{self.id} ({self.action}/{self.player_id})>"



class BaseActionsRoles(MyTable, Base): #, tablename="base_action_roles"
    """Table de données mettant en relation les rôles et les actions de base"""

    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    role = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)
    action = sqlalchemy.Column(sqlalchemy.String(32), nullable=False)

    def __repr__(self):
        """Return repr(self)."""
        return f"<BaseActionsRoles #{self.id} ({self.role}/{self.action})>"



class Taches(MyTable, Base): #, tablename="taches"
    """Table de données des tâches planifiées du bot"""

    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    timestamp = sqlalchemy.Column(sqlalchemy.DateTime(), nullable=False)
    commande = sqlalchemy.Column(sqlalchemy.String(200), nullable=False)
    action = sqlalchemy.Column(sqlalchemy.Integer(), nullable=True)


    def __repr__(self):
        """Return repr(self)."""
        return f"<Taches #{self.id} ({self.commande})>"



class Triggers(MyTable, Base): #, tablename="triggers"
    """Table de données des mots et expressions déclenchant l'IA du bot"""

    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    trigger = sqlalchemy.Column(sqlalchemy.String(500), nullable=False)
    reac_id = sqlalchemy.Column(sqlalchemy.Integer(), nullable=False)


    def __repr__(self):
        """Return repr(self)."""
        extract = self.trigger.replace('\n', ' ')[:15] + "..."
        return f"<Triggers #{self.id} ({extract})>"



class Reactions(MyTable, Base): #, tablename="reactions"
    """Table de données des réactions d'IA connues du bot"""

    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key=True)
    reponse = sqlalchemy.Column(sqlalchemy.String(2000), nullable=False)     # Réponse, dans le format personnalisé : "txt <||> txt <&&> <##>react"

    def __repr__(self):
        """Return repr(self)."""
        extract = self.reponse.replace('\n', ' ')[:15] + "..."
        return f"<Reactions #{self.id} ({extract})>"



class CandidHaro(MyTable, Base): #, tablename="candid_haro"
    """Table de données des candidatures et haros en cours  #PhilippeCandidHaro"""

    id = sqlalchemy.Column(sqlalchemy.Integer(), primary_key = True)
    player_id = sqlalchemy.Column(sqlalchemy.BigInteger(), nullable=False)
    type = sqlalchemy.Column(sqlalchemy.String(11), nullable=False)

    def __repr__(self):
        """Return repr(self)."""
        return f"<CandidHaro #{self.id} ({self.player_id/self.type})>"



def connect():
    global engine, session

    LGREZ_DATABASE_URI = env.load("LGREZ_DATABASE_URI")
    engine = sqlalchemy.create_engine(LGREZ_DATABASE_URI)           # Moteur SQL : connexion avec le serveur

    # Création des tables si elles n'existent pas déjà
    Base.metadata.create_all(engine)

    # Ouverture de la session
    Session = sessionmaker(bind=engine)
    session = Session()



from discord.ext import commands

from lgrez.blocs.bdd import Joueurs
from lgrez.blocs import tools, bdd, env, gsheets, bdd_tools


# Routine d'inscription (fonction appellée par la commande !co)
async def main(bot, member):
    """Processus d'inscription pour <member>"""

    if Joueurs.query.get(member.id):                                            # Joueur dans la bdd = déjà inscrit
        await tools.private_chan(member).set_permissions(member, read_messages=True, send_messages=True)
        await tools.private_chan(member).send(f"Saloww ! {member.mention} tu es déjà inscrit, viens un peu ici enculé !")
        return
    elif chan := tools.get(member.guild.text_channels, topic=f"{member.id}"):   # Inscription en cours
        await chan.set_permissions(member, read_messages=True, send_messages=True)
        await chan.send(f"Tu as déjà un channel à ton nom, {member.mention}, par ici !")
    else:
        chan = await member.guild.create_text_channel(f"conv-bot-{member.name}",  topic=str(member.id),
                                                      category=tools.channel(member, "CONVERSATION BOT"))
        # Crée le channel "conv-bot-nom" avec le topic "member.id" et au bon endroit

        # await chan.edit(category=tools.channel(member, "CONVERSATION BOT")) #Met le channel au bon endroit (GROS CON DE BOT)
        await chan.set_permissions(member, read_messages=True, send_messages=True)


    ### Mise en mode STFU le temps de l'inscription, si pas déjà fait par l'appel à !co (arrivée nouveau membre)

    if chan.id not in bot.in_stfu:
        bot.in_stfu.append(chan.id)


    ### Récupération nom et renommages

    await chan.send(f"Bienvenue {member.mention} ! Je suis le bot à qui tu auras affaire tout au long de la partie.\nTu es ici sur ton channel privé, auquel seul toi et les MJ ont accès ! C'est ici que tu lanceras toutes les commandes pour voter, te renseigner... mais aussi que les MJ discuteront avec toi en cas de soucis.\n\nPas de panique, je vais tout t'expliquer !")

    await tools.sleep(chan, 5)    # tools.sleep(x) = attend x secondes avec l'indicateur typing...'

    await chan.send(f"""Avant toute chose, finalisons ton inscription !\nD'abord, un point règles nécessaire :\n\n{tools.quote_bloc("En t'inscrivant au Loup-Garou de la Rez, tu garantis vouloir participer à cette édition et t'engages à respecter les règles du jeu. Si tu venais à entraver le bon déroulement de la partie pour une raison ou une autre, les MJ s'autorisent à te mute ou t'expulser du Discord sans préavis.")}""")

    await tools.sleep(chan, 5)    # tools.sleep(x) = attend x secondes avec l'indicateur typing...'

    message = await chan.send(f"""C'est bon pour toi ?\n{tools.ital("(Le bot te demandera souvent confirmation, en t'affichant deux réactions comme ceci. Clique sur ✅ si ça te va, sur ❎ sinon. Tu peux aussi répondre (oui, non, ...) par écrit.)")}""")
    if not await tools.yes_no(bot, message):
        await chan.send(f"Pas de soucis. Si tu changes d'avis ou que c'est un missclick, appelle un MJ aled ({tools.code('@MJ')}).")
        return


    await chan.send(f"Parfait. Je vais d'abord avoir besoin de ton (vrai) prénom, celui par lequel on t'appelle au quotidien. Attention, tout troll sera foudracanonné™ {tools.emoji(chan, 'foudra')}")

    def check_chan(m): #Check que le message soit envoyé par l'utilisateur et dans son channel perso
        return m.channel == chan and m.author != bot.user
    ok = False
    while not ok:
        await chan.send(f"""Quel est ton prénom, donc ?\n{tools.ital("(Répond simplement dans ce channel, à l'aide du champ de texte normal)")}""")
        prenom = await tools.wait_for_message(bot, check=check_chan)

        await chan.send(f"Très bien, et ton nom de famille ?")
        nom_famille = await tools.wait_for_message(bot, check=check_chan)
        nom = f"{prenom.content.title()} {nom_famille.content.title()}"         # .title met en majuscule la permière lettre de chaque mot

        message = await chan.send(f"""Tu me dis donc t'appeller {tools.bold(nom)}. C'est bon pour toi ? Pas d'erreur, pas de troll ?""")
        ok = await tools.yes_no(bot, message)


    await chan.edit(name=f"conv-bot-{nom}")         # Renommage conv
    if not tools.role(member, "MJ") in member.roles:
        await member.edit(nick=nom)                 # Renommage joueur (ne peut pas renommer les MJ)

    await chan.send("Parfait ! Je t'ai renommé(e) pour que tout le monde te reconnaisse, et j'ai renommé cette conversation.")

    await tools.sleep(chan, 5)

    a_la_rez = await tools.yes_no(bot, await chan.send("Bien, dernière chose : habites-tu à la Rez ?"))


    if a_la_rez:
        def sortie_num_rez(m):
            return len(m.content) < 200     # Longueur de chambre de rez maximale
        chambre = (await tools.boucle_message(bot, chan, "Alors, quelle est ta chambre ?", sortie_num_rez, rep_message="Désolé, ce n'est pas un numéro de chambre valide, réessaie...")).content
    else:
        chambre = "XXX (chambre MJ)"

    await chan.send(f"{nom}, en chambre {chambre}... Je t'inscris en base !")

    async with chan.typing():     # Envoi indicateur d'écriture pour informer le joueur que le bot fait des trucs
        # Ajout à la BDD

        joueur = Joueurs(discord_id=member.id, _chan_id=chan.id, nom=member.display_name,
                         chambre=chambre, statut="vivant", role="nonattr", camp="Non attribué",
                         votant_village=True, votant_loups=False, role_actif=False)
        bdd.session.add(joueur)
        bdd.session.commit()

        # Ajout au TDB

        cols = [col for col in bdd_tools.get_cols(Joueurs) if not col.startswith('_')]    # On élimine les colonnes locales

        SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")

        workbook = gsheets.connect(SHEET_ID)    # Tableau de bord
        sheet = workbook.worksheet("Journée en cours")
        values = sheet.get_all_values()         # Liste de liste des valeurs des cellules
        NL = len(values)

        head = values[2]            # Ligne d'en-têtes (noms des colonnes) = 3e ligne du TDB
        TDB_index = {col: head.index(col) for col in cols}    # Dictionnaire des indices des colonnes GSheet pour chaque colonne de la table
        TDB_tampon_index = {col: head.index(f"tampon_{col}") for col in cols if col != 'discord_id'}    # Idem pour la partie « tampon »

        plv = 3        # Première ligne vide (si tableau vide, 4e ligne ==> l=3)
        for l in range(NL):
            if values[l][TDB_index["discord_id"]].isdigit():    # Si il y a un vrai ID dans la colonne ID, ligne l
                plv = l + 1

        Modifs = [(plv, TDB_index[col], getattr(joueur, col)) for col in TDB_index] + [(plv, TDB_tampon_index[col], getattr(joueur, col)) for col in TDB_tampon_index]   # Modifs : toutes les colonnes de la partie principale + du cache
        gsheets.update(sheet, Modifs)


        # Grant accès aux channels joueurs et information

        await member.add_roles(tools.role(member, "Joueur en vie"))
        await chan.edit(topic="Ta conversation privée avec le bot, c'est ici que tout se passera !")


    # Conseiller d'ACTIVER TOUTES LES NOTIFS du chan (et mentions only pour le reste, en activant @everyone)
    await chan.send("Tu es maintenant inscrit(e) ! Je t'ai attribué le rôle \"Joueur en vie\", qui te donne l'accès à tout plein de nouveaux channels à découvrir.")
    await chan.send(f"Juste quelques dernières choses :\n - Plein de commandes te sont d'ores et déjà accessibles ! Découvre les toutes en tapant {tools.code('!help')} ;\n - Si tu as besoin d'aide, plus de bouton MJ ALED : mentionne simplement les MJs ({tools.code('@MJ')}) et on viendra voir ce qui se passe !\n - Si ce n'est pas le cas, je te conseille fortement d'installer Discord sur ton téléphone, et d'{tools.bold('activer toutes les notifications')} pour ce channel ! Promis, pas de spam :innocent:\nPour le reste du serveur, tu peux le mettre en \"mentions only\", en activant le {tools.code('@everyone')} – il est limité ;\n\nEnfin, n'hésite pas à me parler, j'ai toujours quelques réponses en stock...")

    await tools.sleep(chan, 5)
    await chan.send("Voilà, c'est tout bon ! Installe toi bien confortablement, la partie commence le 32 plopembre.")

    # Log
    await tools.log(member, f"Inscription de {member.name}#{member.discriminator} réussie\n - Nom : {nom}\n - Chambre : {chambre}\n - Channel créé : {chan.mention}")


    # Retrait du mode STFU
    if chan.id in bot.in_stfu:
        bot.in_stfu.remove(chan.id)



from functools import wraps
import asyncio
import datetime
import unidecode
import re

import discord
import discord.utils
from discord.ext import commands

from lgrez.blocs import bdd, bdd_tools
from lgrez.blocs.bdd import Tables, Joueurs, Roles, BaseActions, Actions, BaseActionsRoles, Taches, Triggers, Reactions, CandidHaro
# on importe toutes les tables, plus simple pour y accéder depuis des réactions etc (via eval_accols)


### ---------------------------------------------------------------------------
### Utilitaires de récupération d'objets Discord (détectent les mentions)
### ---------------------------------------------------------------------------

# Raccourci : tools.get = discord.utils.get
get = discord.utils.get


def find_by_mention_or_name(collec, val, pattern=None, must_be_found=False, raiser=None):
    """Utilitaire pour la suite : trouve <val> dans <collec>

    [pattern]           Motif RegEx à utiliser pour la recherche
    [must_be_found]     Si True, raise une AssertionError si <val> est introuvable
    [raiser]            Nom de la fonction à envoyer dans l'exception si introuvable
    """
    if not val:
        item = None
    elif pattern and (match := re.search(pattern, val)):
        item = get(collec, id=int(match.group(1)))
    else:
        item = get(collec, name=val)

    if must_be_found:
        assert item, f"{raiser or 'tools.find_by_mention_or_name'} : Élément {val} introuvable"

    return item


def channel(arg, nom, must_be_found=True):
    """Renvoie l'objet discord.Channel du channel #<nom>.

    <nom>               Nom du channel (texte/vocal/catégorie) ou sa mention (détection directe par RegEx)
    <arg>               Argument permettant de remonter aux channels : discord.Context, discord.Guild, discord.Member ou discord.Channel
    [must_be_found]     Si True (défaut), raise une AssertionError si le channel #<nom> n'existe pas (si False, renvoie None)
    """
    try:
        channels = arg.channels if isinstance(arg, discord.Guild) else arg.guild.channels
    except AttributeError:
        raise TypeError("tools.channel : Impossible de remonter aux channels depuis l'argument trasmis")
    return find_by_mention_or_name(channels, nom, pattern="<#([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.channel")


def role(arg, nom, must_be_found=True):
    """Renvoie l'objet discord.Role du rôle @&nom.

    <nom>               Nom du rôle ou sa mention (détection directe par RegEx)
    <arg>               Argument permettant de remonter aux rôles : discord.Context, discord.Guild, discord.Member ou discord.Channel
    [must_be_found]     Si True (défaut), raise une AssertionError si le rôle @&nom n'existe pas (si False, renvoie None)
    """
    try:
        roles = arg.roles if isinstance(arg, discord.Guild) else arg.guild.roles
    except AttributeError:
        raise TypeError("tools.role : Impossible de remonter aux rôles depuis l'argument trasmis")
    return find_by_mention_or_name(roles, nom, pattern="<@&([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.role")


def member(arg, nom, must_be_found=True):
    """Renvoie l'objet discord.Member du membre @member.

    <nom>               Nom du joueur ou sa mention (détection directe par RegEx)
    <arg>               Argument permettant de remonter aux rôles : discord.Context, discord.Guild, discord.Member ou discord.Channel
    [must_be_found]     Si True (défaut), raise une AssertionError si le membre @member n'existe pas (si False, renvoie None)
    """
    try:
        members = arg.members if isinstance(arg, discord.Guild) else arg.guild.members
    except AttributeError:
        raise TypeError("tools.member : Impossible de remonter aux membres depuis l'argument trasmis")
    return find_by_mention_or_name(members, nom, pattern="<@!([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.member")


def emoji(arg, nom, must_be_found=True):
    """Renvoie l'objet discord.Emoji de l'emoji :nom:.

    <nom>               Nom de l'emoji ou son utilisation (détection directe par RegEx)
    <arg>               Argument permettant de remonter aux rôles : discord.Context, discord.Guild, discord.Member ou discord.Channel
    [must_be_found]     Si True (défaut), raise une AssertionError si l'emoji :nom: n'existe pas (si False, renvoie None)
    """
    try:
        emojis = arg.emojis if isinstance(arg, discord.Guild) else arg.guild.emojis
    except AttributeError:
        raise TypeError("tools.emoji : Impossible de remonter aux emojis depuis l'argument trasmis")
    return find_by_mention_or_name(emojis, nom, pattern="<:.*:([0-9]{18})>",
                                   must_be_found=must_be_found, raiser="tools.emoji")


def private_chan(member, must_be_found=True):
    """Renvoie le channel privé de <member> (type discord.Member)

    [must_be_found]     Si True (défaut), raise une AssertionError si le channel n'existe pas (si False, renvoie None)
    """
    joueur = Joueurs.query.get(member.id)
    assert joueur, f"tools.private_chan : Joueur {member} introuvable"
    chan = member.guild.get_channel(joueur._chan_id)
    if must_be_found:
        assert chan, f"tools.private_chan : Chan privé de {joueur} introuvable"
    return chan


# Appel aux MJs
def mention_MJ(arg):
    """Renvoie @MJ si le joueur n'est pas un MJ.

    <arg> peut être de type discord.Context ou discord.Member
    """
    member = arg.author if hasattr(arg, "author") else arg
    if hasattr(member, "top_role") and member.top_role.name == "MJ":    # Si webhook, pas de top_role
        return "@MJ"
    else:
        return role(arg, "MJ").mention



### ---------------------------------------------------------------------------
### Exceptions
### ---------------------------------------------------------------------------

class CommandExit(RuntimeError):
    """Force l'arrêt immédiat d'une commande, et empêche le bot de réagir à nouveau"""

    pass


### ---------------------------------------------------------------------------
### Décorateurs pour les différentes commandes, en fonction de leur usage
### ---------------------------------------------------------------------------

# @tools.mjs_only : commande exécutables uniquement par un MJ ou un webhook
mjs_only = commands.check_any(commands.check(lambda ctx: ctx.message.webhook_id), commands.has_any_role("MJ", "Bot"))

# @tools.mjs_et_redacteurs : commande exécutables par un MJ, un rédacteur ou un webhook (pour IA)
mjs_et_redacteurs = commands.check_any(commands.check(lambda ctx: ctx.message.webhook_id), commands.has_any_role("MJ", "Bot", "Rédacteur"))

# @tools.joueurs_only : commande exécutables uniquement par un joueur (inscrit en base), vivant ou mort
joueurs_only = commands.has_any_role("Joueur en vie", "Joueur mort")

# @tools.vivants_only : commande exécutables uniquement par un joueur vivant
vivants_only = commands.has_role("Joueur en vie")

# @tools.private : utilisable en combinaison avec joueurs_only et vivants_only (pas avec les autres attention, vu que seuls les joueurs ont un channel privé)
def private(cmd):
    """Supprime le message et exécute la commande dans la conv privée si elle a été appellée ailleurs.

    Ce décorateur n'est utilisable que sur une commande définie dans un Cog.
    Si le joueur ayant utilisé la commande n'a pas de chan privé (pas en base), raise une AssertionError.
    """
    @wraps(cmd)
    async def new_cmd(self, ctx, *args, **kwargs):              # Cette commande est renvoyée à la place de cmd
        if not ctx.channel.name.startswith("conv-bot-"):            # Si pas déjà dans une conv bot :
            await ctx.message.delete()                                  # On supprime le message,
            ctx.channel = private_chan(ctx.author)                      # On remplace le chan dans le contexte d'appel par le chan privé,
            await ctx.send(f"{quote(ctx.message.content)}\n"            # On envoie un warning dans le chan privé,
                           f"{ctx.author.mention} :warning: Cette commande est interdite en dehors de ta conv privée ! :warning:\n"
                           f"J'ai supprimé ton message, et j'exécute la commande ici :")
        return await cmd(self, ctx, *args, **kwargs)                # Et on appelle cmd, avec le contexte modifié !

    return new_cmd



### ---------------------------------------------------------------------------
### Commandes d'interaction avec les joueurs : input, boucles, confirmation...
### ---------------------------------------------------------------------------

# Commande générale, à utiliser à la place de bot.wait_for('message', ...)
async def wait_for_message(bot, check, trigger_on_commands=False):
    """Attend et renvoie le premier message reçu rencontrant les critères demandés.

    Surcouche de bot.wait_for() permettant d'ignoer les commandes et de réagir au mot-clé STOP :
    <check> fonction discord.Message -> bool
    [trigger_on_commands]   Si False (défaut), un message respectant <check> sera ignoré si c'est une commande

    Si le message est "stop" ou "!stop" (ou autre casse), raise une exception CommandExit (même si le message respecte <check>).
    """
    if trigger_on_commands:
        def trig_check(m):
            return (check(m) or m.content.lower() in ["stop", "!stop"])         # et on trigger en cas de STOP
    else:
        def trig_check(m):
            return ((check(m)
                     and not m.content.startswith(bot.command_prefix))          # on ne trigger pas sur les commandes
                    or m.content.lower() in ["stop", "!stop"])                  # et on trigger en cas de STOP

    message = await bot.wait_for('message', check=trig_check)
    if message.content.lower() in ["stop", "!stop"]:
        raise CommandExit("Arrêt demandé")
    else:
        return message


# Raccourci pratique
async def wait_for_message_here(ctx, trigger_on_commands=False):
    """Attend et renvoie le premier message reçu dans <ctx>.

    Surcouche de wait_for_message filtrant uniquement les messages envoyés dans <ctx>.channel par quelqu'un d'autre que le bot
    [trigger_on_commands]   Passé directement à wait_for_message
    """
    def trig_check(message):
        return (message.channel == ctx.channel and message.author != ctx.bot.user)

    message = await wait_for_message(ctx.bot, check=trig_check, trigger_on_commands=trigger_on_commands)
    return message


# Permet de boucler question -> réponse tant que la réponse vérifie pas les critères nécessaires dans chan
async def boucle_message(bot, chan, in_message, condition_sortie, rep_message=None):
    """Permet de lancer une boucle question/réponse tant que la réponse ne vérifie pas <condition_sortie>

    <chan>          Channel dans lequel lancer la boucle
    [in_message]    Si défini, message à envoyer avant la boucle
    [rep_message]   Si défini, permet de définir un message de boucle différent de [in_message] (identique si None). Si [in_message] n'est pas défini, doit être défini.
    """
    if not rep_message:
        rep_message = in_message
    if not rep_message:
        raise ValueError("tools.boucle_message : [in_message] ou [rep_message] doit être défini !")

    def check_chan(m): #C heck que le message soit envoyé par l'utilisateur et dans son channel perso
        return m.channel == chan and m.author != bot.user

    await chan.send(in_message)
    rep = await wait_for_message(bot, check_chan)
    while not condition_sortie(rep):
        await chan.send(rep_message)
        rep = await wait_for_message(bot, check_chan)

    return rep


async def boucle_query_joueur(ctx, cible=None, message=None, sensi=0.5):
    """Retourne un joueur dans le contexte <ctx>.

    [cible]     Cible par défaut (donnée par le joueur dès le début)
    [message]   Si défini (et [cible] non définie), message à envoyer avant la boucle
    [sensi]     Sensibilité de la recherche (défaut 0.5)

    Attend que le joueur entre un nom de joueur, et boucle 5 fois au max (avant de l'insulter et de raise une erreur) pour chercher le plus proche joueurs dans la table Joueurs.
    """
    if message and not cible:
        await ctx.send(message)

    for i in range(5):
        if i == 0 and cible:            # Au premier tour, si on a donné une cible
            rep = cible
        else:
            mess = await wait_for_message_here(ctx)
            rep = mess.content

        if id := ''.join([c for c in rep if c.isdigit()]):      # Si la chaîne contient un nombre, on l'extrait
            if joueur := Joueurs.query.get(int(id)):                # Si cet ID correspond à un utilisateur, on le récupère
                return joueur                                       # On a trouvé l'utilisateur !

        nearest = await bdd_tools.find_nearest(rep, Joueurs, carac="nom", sensi=sensi)     # Sinon, recherche au plus proche

        if not nearest:
            await ctx.send("Aucune entrée trouvée, merci de réessayer :")

        elif nearest[0][1] == 1:        # Si le score le plus haut est égal à 1...
            return nearest[0][0]        # ...renvoyer l'entrée correspondante

        elif len(nearest) == 1:
            m = await ctx.send(f"Je n'ai trouvé qu'une correspondance : {nearest[0][0].nom}\nÇa part ?")
            if await yes_no(ctx.bot, m):
                return nearest[0][0]
            else:
                await ctx.send("Bon d'accord, alors tu votes contre qui ?")

        else:
            s = "Les joueurs les plus proches de ton entrée sont les suivants : \n"
            for i, j in enumerate(nearest[:10]):
                s += f"{emoji_chiffre(i+1)}. {j[0].nom} \n"
            m = await ctx.send(s + tools.ital("Tu peux les choisir en réagissant à ce message, ou en répondant au clavier."))
            n = await choice(ctx.bot, m, min(10, len(nearest)))
            return nearest[n-1][0]

    await ctx.send("Et puis non, tiens ! \n https://giphy.com/gifs/fuck-you-middle-finger-ryan-stiles-x1kS7NRIcIigU")
    raise RuntimeError("Le joueur est trop con, je peux rien faire")


# Récupère un input par réaction
async def wait_for_react_clic(bot, message, emojis={}, *, process_text=False,
                              text_filter=lambda s: True, post_converter=None, trigger_all_reacts=False, trigger_on_commands=False):
    """Ajoute les reacts dans [emojis] à message, attend que quelqu'un appuie sur une, puis renvoie :
        - soit le nom de l'emoji si [emojis] est une liste ;
        - soit la valeur associée si [emojis] est un dictionnaire.

    Si [process_text] == True, détecte aussi la réponse par message et retourne ledit message (défaut False).
    De plus, si [text_filter] (fonction str -> bool) est défini, ne réagit qu'aux messages pour lesquels text_filter(message) = True.
    De plus, si [post_converter] (fonction str -> ?) est défini, le message détecté est passé dans cette fonction avant d'être renvoyé.

    Si [trigger_all_reacts] == True, détecte l'ajout des toutes les réactions (et pas seulement celles dans [emojis]) et renvoie l'emoji directement si il n'est pas dans [emojis] (défaut False).
    Enfin, [trigger_on_commands] est passé directement à wait_for_message.
    """

    if not isinstance(emojis, dict):        # Si emoji est une liste, on en fait un dictionnaire
        emojis = {emoji: emoji for emoji in emojis}

    try:    # Si une erreur dans ce bloc, on supprime les emojis du bot (sinon c'est moche)
        for emoji in emojis:                    # On ajoute les emojis
            await message.add_reaction(emoji)

        emojis_names = [emoji.name if hasattr(emoji, "name") else emoji for emoji in emojis]
        def react_check(react):                     # Check REACT : bon message, pas un autre emoji, et pas react du bot
            return (react.message_id == message.id
                    and react.user_id != bot.user.id
                    and (trigger_all_reacts or react.emoji.name in emojis_names))

        react_task = asyncio.create_task(bot.wait_for('raw_reaction_add', check=react_check), name="react")

        if process_text:
            def message_check(mess):        # Check MESSAGE : bon channel, pas du bot, et filtre
                return (mess.channel == message.channel
                        and mess.author != bot.user
                        and text_filter(mess.content))
        else:
            def message_check(mess):        # On process DANS TOUS LES CAS, mais juste pour détecter "stop" si process_text == False
                return False

        mess_task = asyncio.create_task(wait_for_message(bot, check=message_check, trigger_on_commands=True), name="mess")

        done, pending = await asyncio.wait([react_task, mess_task], return_when=asyncio.FIRST_COMPLETED)      # On lance
        # Le bot attend ici qu'une des deux tâches aboutissent
        done_task = list(done)[0]        # done = tâche réussie

        if done_task.get_name() == "react":
            emoji = done_task.result().emoji
            if trigger_all_reacts and emoji.name not in emojis_names:
                ret = emoji
            else:
                ret = emojis[emoji.name]                            # Si clic sur react, done.result = react

            for emoji in emojis:
                await message.remove_reaction(emoji, bot.user)      # On finit par supprimer les emojis mis par le bot

        else:   # Réponse par message / STOP
            mess = done_task.result().content                # Si envoi de message, done.result = message
            ret = post_converter(mess) if post_converter else mess
            await message.clear_reactions()

    except Exception as exc:
        await message.clear_reactions()
        raise exc from Exception

    return ret


# Surcouche de wait_for_react_clic pour demander une confirmation / question fermée simplement
async def yes_no(bot, message):
    """Ajoute les reacts ✅ et ❎ à message et renvoie True ou False en fonction de l'emoji cliqué OU de la réponse textuelle détectée."""
    yes_words = ["oui", "o", "yes", "y", "1", "true"]
    yes_no_words = yes_words + ["non", "n", "no", "n", "0", "false"]
    return await wait_for_react_clic(
        bot, message, emojis={"✅": True, "❎": False}, process_text=True,
        text_filter=lambda s: s.lower() in yes_no_words, post_converter=lambda s: s.lower() in yes_words)


# Surcouche de wait_for_react_clic pour demander de choisir dans une liste simplement
async def choice(bot, message, N, start=1):
    """Ajoute les reacts [start]=1️⃣, 2️⃣, 3️⃣... <N> à message et renvoie le numéro cliqué OU détecté par réponse textuelle. (N <= 10)"""
    return await wait_for_react_clic(
        bot, message, emojis={emoji_chiffre(i): i for i in range(start, N+1)}, process_text=True,
        text_filter=lambda s: s.isdigit() and start <= int(s) <= N, post_converter=int)


async def sleep(chan, x):
    """Attend <x> secondes en affichant l'indicateur typing... sur <chan>"""
    async with chan.typing():
        await asyncio.sleep(x)



### ---------------------------------------------------------------------------
### Utilitaires d'emojis
### ---------------------------------------------------------------------------

def montre(heure=None):
    """Renvoie l'emoji horloge correspondant à l'heure demandée.

    [heure] str "XXh" ou "XXh30", actuelle si non précisée
    """
    if heure and isinstance(heure, str):
        heure, minute = heure.split("h")
        heure = int(heure) % 12
        minute = int(minute) % 60 if minute else 0
    else:
        now = datetime.datetime.now()
        heure = now.hour % 12
        minute = now.minute

    if 15 < minute < 45:        # Demi heure
        L = ["🕧", "🕜", "🕝", "🕞", "🕟", "🕠", "🕡", "🕢", "🕣", "🕤", "🕥", "🕦"]
    else:                       # Heure pile
        L = ["🕛", "🕐", "🕑", "🕒", "🕓", "🕔", "🕕", "🕖", "🕗", "🕘", "🕙", "🕚"]
    return L[heure] if minute < 45 else L[(heure + 1) % 12]


def emoji_chiffre(chiffre, multi=False):
    """Renvoie l'emoji 0️⃣, 1️⃣, 2️⃣... correspond à <chiffre>.

    Si [multi] == True, <chiffre> doit être un entier positif dont les chiffres seront convertis séparément.
    Sinon (par défaut), <chiffre> doit être un entier entre 0 et 10.
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 10:
        return ["0️⃣", "1️⃣", "2️⃣", "3️⃣", "4️⃣", "5️⃣", "6️⃣", "7️⃣", "8️⃣", "9️⃣", "🔟"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([emoji_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.emoji_chiffre doit être un entier entre 0 et 10 OU un entier positif avec multi=True")


def super_chiffre(chiffre, multi=False):
    """Renvoie le caractère unicode ⁰, ¹, ²... correspond à <chiffre>.

    Si [multi] == True, <chiffre> doit être un entier positif dont les chiffres seront convertis séparément.
    Sinon (par défaut), <chiffre> doit être un entier entre 0 et 9.
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["⁰", "¹", "²", "³", "⁴", "⁵", "⁶", "⁷", "⁸", "⁹"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([super_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.super_chiffre doit être un entier entre 0 et 9 OU un entier positif avec multi=True")


def sub_chiffre(chiffre: int, multi=False):
    """Renvoie le caractère unicode ₀, ₁, ₂... correspond à <chiffre>.

    Si [multi] == True, <chiffre> doit être un entier positif dont les chiffres seront convertis séparément.
    Sinon (par défaut), <chiffre> doit être un entier entre 0 et 9.
    """
    if isinstance(chiffre, int) and 0 <= chiffre <= 9:
        return ["₀", "₁", "₂", "₃", "₄", "₅", "₆", "₇", "₈", "₉"][chiffre]
    elif multi and str(chiffre).isdigit():
        return ''.join([sub_chiffre(int(c)) for c in str(chiffre)])
    else:
        raise ValueError("L'argument de tools.sub_chiffre doit être un entier entre 0 et 9 OU un entier positif avec multi=True")



### ---------------------------------------------------------------------------
### Utilitaires de date / temps, notemment liées aux horaires de jeu
### ---------------------------------------------------------------------------

def heure_to_time(heure):
    """Convertit <heure> = HHh[MM] (str) en objet datetime.time."""
    try:
        hh, mm = heure.split("h")
        return datetime.time(int(hh), int(mm) if mm else 0)
    except ValueError as exc:
        raise ValueError(f"Valeur \"{heure}\" non convertible en temps") from exc


def time_to_heure(tps, sep="h", force_minutes=False):
    """Convertit <tps> (objet datetime.time) en str "HH[sep]" / "HH[sep]MM".

    [sep]               séparateur heures / minutes (défaut "h")
    [force_minutes]     si False (défaut), les minutes ne sont indiquées que si différentes de 0.

    Renvoit une chaîne vide si <tps> est None.
    """
    if tps:
        sep = sep.replace("%", "%%")    # Échappement des % pour utilisation dans strftime

        if force_minutes or tps.minute > 0:
            return f"{tps.hour}{sep}{tps.minute:02}"
        else:
            return f"{tps.hour}{sep}"
    else:
        return ""


# Renvoie le datetime correspondant au prochain moment ou tps arrive DANS LES HORAIRES DU JEU : du dimanche 19:00:00 au vendredi 18:59:59.
def next_occurence(tps):
    """Renvoie l'objet datetime.datetime correspondant à la prochaine occurence de <tps> dans le cadre du jeu.

    <tps> objet datetime.time.
    Renvoie le prochain timestamp arrivant DANS LES HORAIRES DU JEU : du dimanche 19:00:00 au vendredi 18:59:59.
    """
    pause = datetime.time(hour=19)

    now = datetime.datetime.now()
    jour = now.date()
    if tps <= now.time():       # Si plus tôt dans la journée que l'heure actuelle
        jour += datetime.timedelta(days=1)       # on réfléchit comme si on était demain très tôt

    wd = jour.weekday()         # Jour de la semaine, Lu = 0 ... Di = 6

    if tps < pause:
        if wd <= 4:                 # Avant 19h du lundi au vendredi : OK
            pass
        else:                       # Avant 19h mais on est samedi/dimanche
            jour += datetime.timedelta(days=(7-wd))
    else:
        if wd <= 3 or wd == 6:      # Après 19h du dimanche au jeudi : OK
            pass
        else:                       # Après 19h et on est vendredi/samedi
            jour += datetime.timedelta(days=(6-wd))

    return datetime.datetime.combine(jour, tps)         # passage de date et time à datetime


def debut_pause():
    """Renvoie l'objet datetime.datetime correspondant au prochain vendredi 19h."""
    pause_time = datetime.time(hour=19)
    pause_wday = 4          # Vendredi

    now = datetime.datetime.now()
    jour = now.date()
    if pause_time <= now.time():        # Si plus tôt dans la journée que l'heure actuelle
        jour += datetime.timedelta(days=1)       # on réfléchit comme si on était demain très tôt

    pause_jour = jour + datetime.timedelta(days=(pause_wday - jour.weekday()) % 7)      # Jour décalé du nombre de jours avant vendredi
    return datetime.datetime.combine(pause_jour, pause_time)         # passage de date et time à datetime


def fin_pause():
    """Renvoie l'objet datetime.datetime correspondant au prochain dimanche 19h."""
    reprise_time = datetime.time(hour=19)
    reprise_wday = 6        # Dimanche

    now = datetime.datetime.now()
    jour = now.date()
    if reprise_time <= now.time():      # Si plus tôt dans la journée que l'heure actuelle
        jour += datetime.timedelta(days=1)      # on réfléchit comme si on était demain très tôt

    reprise_jour = jour + datetime.timedelta(days=(reprise_wday - jour.weekday()) % 7)      # Jour décalé du nombre de jours avant vendredi
    return datetime.datetime.combine(reprise_jour, reprise_time)        # passage de date et time à datetime



### ---------------------------------------------------------------------------
### Split et log
### ---------------------------------------------------------------------------

# Sépare <mess> en une liste de messages de moins de <N>=2000 mots (limitation Discord), en séparant aux <sep>=sauts de ligne si possible.
# Ajoute <rep> à la fin des messages tronqués de leur séparateur final.
def smooth_split(mess, N=1990, sep='\n', rep=''):
    """Sépare <mess> en une liste de messages de moins de [N]=1990 mots.

    [sep]   Caractères où séparer préférentiellement le texte (défaut sauts de ligne). Si <message> contient une sous-chaîne plus longue que [N] ne contenant pas [sep], le message est tronqué à la limite
    <rep>   Chaîne ajoutée à la fin de chaque message (tronqué du séparateur final)

    1990 car 2000 est la limitation Discord, et on laisse de la marge (typiquement si dans un bloc code, +6 caractères)
    """
    mess = str(mess)
    LM = []             # Liste des messages
    psl = 0             # indice du Précédent Saut de Ligne
    L = len(mess)
    while psl + N < L:
        if mess.count(sep, psl, psl+N+len(sep)):       # +len(sep) parce que si sep est à la fin, on le dégage
            i = psl + N - mess[psl: psl+N+len(sep)][::-1].find(sep)      # un peu sombre mais vrai, tkt frère
            LM.append(mess[psl: i] + rep)
            psl = i + 1     # on élimine le \n
        else:
            LM.append(mess[psl: psl + N])
            psl += N

    if psl < L:
        LM.append(mess[psl:])   # ce qui reste
    return LM


async def send_blocs(messageable, mess, N=1990, sep='\n', rep=''):
    """Envoie <mess> dans <messageable> (ctx / channel)"""
    [await messageable.send(bloc) for bloc in smooth_split(mess, N=N, sep=sep, rep=rep)]


async def send_code_blocs(messageable, mess, N=1990, sep='\n', rep='', langage=""):
    """Envoie dans <messageable> (ctx / channel) <mess> sous forme de blocs de code"""
    [await messageable.send(code_bloc(bloc, langage=langage)) for bloc in smooth_split(mess, N=N, sep=sep, rep=rep)]


# Log dans #logs
async def log(arg, message, code=False):
    """Envoie <message> dans le channel #logs.

    <arg>       Argument permettant de remonter aux rôles : discord.Context, discord.Guild, discord.Member ou discord.Channel
    [code]      Si True, log sous forme de bloc(s) de code (défaut False)
    """
    logchan = channel(arg, "logs")
    if code:
        await send_code_blocs(logchan, message)
    else:
        [await logchan.send(bloc) for bloc in smooth_split(message)]



### ---------------------------------------------------------------------------
### Autres fonctions diverses
### ---------------------------------------------------------------------------

# Crée un contexte à partir d'un message_id : simule que <member> a envoyé <content> dans son chan privé
async def create_context(bot, message_id, member, content):
    """Renvoie un objet contexte de commande (objet discord.ext.commands.Context) à partir de <message_id>

    Simule que <member> a envoyé <content> dans son chan privé et "génère" le contexte associé
    <member> doit être un joueur inscrit en base (pour avoir un chan privé)
    """
    chan = private_chan(member)
    message = (await chan.history(limit=1).flatten())[0]        # On a besoin de récupérer un message, ici le dernier de la conv privée
    # message = await chan.fetch_message(message_id)
    message.author = member
    message.content = content
    return await bot.get_context(message)


def nom_role(role, prefixe=False):
    """Retourne le nom du slug <role> (None si non trouvé)"""
    if role := Roles.query.get(role):
        if prefixe:
            return f"{role.prefixe}{role.nom}"
        else:
            return role.nom
    else:
        return None


# Remove accents
def remove_accents(s):
    """Renvoie la chaîne non accentuée, mais conserve les caractères spéciaux (emojis...)"""
    p = re.compile("([À-ʲΆ-ת])")      # Abracadabrax, c'est moche mais ça marche (source : tkt frère)
    return p.sub(lambda c: unidecode.unidecode(c.group()), s)


# Évaluation d'accolades
def eval_accols(rep, globals=None, locals=None, debug=False):
    """Replace chaque bloc entouré par des {} par leur évaluation Python.

    [globals]   Dictionnaire des variables globales du contexte d'évaluation (passé à eval)
    [locals]    Dictionnaire des variables locales du contexte d'évaluation (passé à eval)
    [debug]     Si False (défaut), laisse l'expression telle quelle (avec les accolades) si une exception est levée durant l'évaluation.
                Si True, insère le message d'erreur (type et texte de l'exception dans le message) ensuite.

    Penser à passer les globals() et locals() si besoin. Généralement, il faut passer locals() qui contient ctx, etc... mais pas globals() si on veut bénéficier de tous les modules importés dans tools.py.
    """
    if "{" in rep:              # Si contient des expressions
        evrep = ""                  # Réponse évaluée
        expr = ""                   # Expression à évaluer
        noc = 0                     # Nombre de { non appariés
        for c in rep:
            if c == "{":
                if noc:             # Expression en cours :
                    expr += c           # on garde le {
                noc += 1
            elif c == "}":
                noc -= 1
                if noc:             # idem
                    expr += c
                else:               # Fin d'une expression
                    try:                                            # On essaie d'évaluer la chaîne
                        evrep += str(eval(expr, globals, locals))       # eval("expr") = expr
                    except Exception as e:
                        evrep += "{" + expr + "}"                   # Si erreur, on laisse {expr} non évaluée
                        if debug:
                            evrep += tools.code(f"->!!! {e} !!!")
                    expr = ""
            elif noc:               # Expression en cours
                expr += c
            else:                   # Pas d'expression en cours
                evrep += c
        if noc:     # Si expression jamais finie (nombre impair de {)
            evrep += "{" + expr
        return evrep
    else:
        return rep



### ---------------------------------------------------------------------------
### Utilitaires de formatage de texte
### ---------------------------------------------------------------------------

def bold(s):
    """Retourne <s> formaté comme texte en gras dans Discord"""
    return f"**{s}**"

def ital(s):
    """Retourne <s> formaté comme texte en italique dans Discord"""
    return f"*{s}*"

def soul(s):
    """Retourne <s> formaté comme texte souligné dans Discord"""
    return f"__{s}__"

def strike(s):
    """Retourne <s> formaté comme texte barré dans Discord"""
    return f"~~{s}~~"

def code(s):
    """Retourne <s> formaté comme code (inline) dans Discord"""
    return f"`{s}`"

def code_bloc(s, langage=""):
    """Retourne <s> formaté comme un bloc de code dans Discord

    [langage]  langage du code, permet la coloration syntaxique (ordinateur uniquement).
    Langages supportés (non exhaustif ?) : asciidoc, autohotkey, bash, coffeescript, cpp (C++), cs (C#), css, diff, fix, glsl, ini, json, md, (markdown), ml, prolog, py, tex, xl, xml
    """
    return f"```{langage}\n{s}```"

def quote(s):
    """Retourne <s> formaté comme citation (inline) dans Discord"""
    return f"> {s}"

def quote_bloc(s):
    """Retourne <s> formaté comme bloc de citation (multiline) dans Discord"""
    return f">>> {s}"

def spoiler(s):
    """Retourne <s> formaté comme spoiler (cliquer pour afficher) dans Discord"""
    return f"||{s}||"



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.readlines()

setuptools.setup(
    name="lg-rez",
    version="1.0.2",
    author="Loïc Simon, Tom Lacoma",
    author_email="loic.simon@espci.org, tom.lacoma@espci.org",
    description="Discord bot for organizing Werewolf RP games ESPCI-style",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/loic-simon/lg-rez",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements,
    python_requires='>=3.8',
)



import datetime

from discord.ext import commands

from lgrez.blocs import env, webhook, bdd, tools
from lgrez.blocs.bdd import Taches, Actions, Joueurs



def execute(tache):
    """Exécute la tâche <tache> (objet BDD Taches) : appelle le webhook et nettoie"""
    LGREZ_WEBHOOK_URL = env.load("LGREZ_WEBHOOK_URL")
    webhook.send(tache.commande, url=LGREZ_WEBHOOK_URL)
    bdd.session.delete(tache)
    bdd.session.commit()


def add_task(bot, timestamp, commande, action=None):
    """Ajoute une tâche sur le bot + en base (fonction pour usage ici et dans d'autres features)"""
    now = datetime.datetime.now()
    tache = Taches(timestamp=timestamp, commande=commande, action=action)

    bdd.session.add(tache)                                                           # Enregistre la tâche en BDD
    bdd.session.commit()

    TH = bot.loop.call_later((timestamp - now).total_seconds(), execute, tache)     # Programme la tâche (appellera execute(tache) à timestamp)
    bot.tasks[tache.id] = TH        # TaskHandler, pour pouvoir cancel


def cancel_task(bot, tache):
    """Supprime (annule) une tâche (fonction pour usage ici et dans d'autres features)"""
    bot.tasks[tache.id].cancel()        # Annulation (objet TaskHandler)
    bdd.session.delete(tache)            # Suppression en base
    bdd.session.commit()
    del bot.tasks[tache.id]             # Suppression TaskHandler


class GestionTaches(commands.Cog):
    """GestionTaches - Commandes de planification, exécution, annulation de tâches"""

    @commands.command()
    @tools.mjs_only
    async def taches(self, ctx):
        """Liste les tâches en attente (COMMANDE MJ)

        Affiche les commandes en attente d'exécution (dans la table Taches) et le timestamp d'exécution associé.
        Lorsque la tâche est liée à une action, affiche le nom de l'action et du joueur concerné.
        """

        taches = Taches.query.order_by(Taches.timestamp).all()
        LT = ""
        for tache in taches:
            LT += f"\n{str(tache.id).ljust(5)} {tache.timestamp.strftime('%d/%m/%Y %H:%M:%S')}    {tache.commande.ljust(25)} "
            if tache.action and (action := Actions.query.get(tache.action)):
                joueur = Joueurs.query.get(action.player_id)
                assert joueur, f"!taches : joueur d'ID {action.player_id} introuvable"
                LT += f"{action.action.ljust(20)} {joueur.nom}"

        mess = ("Tâches en attente : \n\nID    Timestamp              Commande                  Action               Joueur"
                f"\n{'-'*105}{LT}\n\n"
                "Utilisez !cancel <ID> pour annuler une tâche.") if LT else "Aucune tâche en attente."

        await tools.send_code_blocs(ctx, mess)


    @commands.command(aliases=["doat"])
    @tools.mjs_only
    async def planif(self, ctx, quand, *, commande):
        """Planifie une tâche au moment voulu (COMMANDE MJ)

        - <quand> : format [<J>/<M>[/<AAAA>]-]<H>:<M>[:<S>], avec <J> (jours), <M> (mois), <AAAA> (année sur 4 chiffres), <H> (heures) et <M> (minutes) des entiers et <S> (secondes) un entier ou un flottant, optionnel (défaut : 0).
        La date est optionnelle (défaut : date du jour). Si elle est précisée, elle doit être séparée de l'heure par un tiret et l'année peut être omise (défaut : année actuelle) ;

        - <commande> : commande à exécuter (commençant par un !). La commande sera exécutée PAR UN WEBHOOK et DANS LE CHAN #logs : toutes les commandes qui sont liées au joueur ou réservées au chan privé sont à proscrire (ou doivent a minima être précédées de !doas <cible>)

        Cette commande repose sur l'architecture en base de données, ce qui garantit l'exécution de la commande même si le bot plante entre temps.
        Si le bot est down à l'heure d'exécution prévue, la commande sera exécutée dès le bot de retour en ligne.
        Si la date est dans le passé, la commande est exécutée immédiatement.

        Ex. : - !planif 18:00 !close maire
              - !planif 13/06-10:00 !open maire
              - !planif 13/06/2020-10:00 !open maire
              - !planif 23:25:12 !close maire
        """
        now = datetime.datetime.now()

        if "/" in quand:            # Date précisée
            date, time = quand.split("-")
            J, MA = date.split("/", maxsplit=1)
            day = int(J)
            if "/" in MA:           # Année précisée
                M, A = MA.split("/")
                month = int(M)
                year = int(A)
            else:
                month = int(MA)
                year = now.year
            date = datetime.date(year=year, month=month, day=day)
        else:
            date = now.date()
            time = quand

        H, MS = time.split(":", maxsplit=1)
        hour = int(H)
        if ":" in MS:               # Secondes précisées
            M, S = MS.split(":")
            minute = int(M)
            second = int(S)
        else:
            minute = int(MS)
            second = 0
        time = datetime.time(hour=hour, minute=minute, second=second)

        ts = datetime.datetime.combine(date, time)
        message = await ctx.send(f"Planifier {tools.code(commande)} pour le {tools.code(ts.strftime('%d/%m/%Y %H:%M:%S'))} ?")

        if await tools.yes_no(ctx.bot, message):
            action_id = None            # ID de l'action associée à la tâche (utile pour propagation à la suppression de l'action)
            try:
                quoi, id = commande.split(" ")
                if quoi in ["!open", "!close", "!remind"]:
                    action_id = int(id)
            except ValueError:
                pass

            add_task(ctx.bot, ts, commande, action=action_id)
            await ctx.send("Fait.")

        else:
            await ctx.send("Mission aborted.")


    @commands.command(aliases=["retard", "doin"])
    @tools.mjs_only
    async def delay(self, ctx, duree, *, commande):
        """Exécute une commande après XhYmZs (COMMANDE MJ)

        - <duree> : format [<X>h][<Y>m][<Z>s], avec <X> (heures) et <Y> (minutes) des entiers et <Z> (secondes) un entier ou un flottant. Chacune des trois composantes est optionnelle, mais au moins une d'entre elle doit être présente ;
        - <commande> : commande à exécuter (commençant par un !). La commande sera exécutée PAR UN WEBHOOK et DANS LE CHAN #logs : toutes les commandes qui sont liées au joueur ou réservées au chan privé sont à proscrire (ou doivent a minima être précédées de !doas <cible>)

        Cette commande repose sur l'architecture en base de données, ce qui garantit l'exécution de la commande même si le bot plante entre temps.
        Si le bot est down à l'heure d'exécution prévue, la commande sera exécutée dès le bot de retour en ligne.

        Ex. : - !delay 2h !close maire
              - !delay 1h30m !doas @moi !vote Yacine Oussar
        """
        secondes = 0
        try:
            if "h" in duree.lower():
                h, duree = duree.split("h")
                secondes += 3600*int(h)
            if "m" in duree.lower():
                m, duree = duree.split("m")
                secondes += 60*int(m)
            if "s" in duree.lower():
                s, duree = duree.split("s")
                secondes += float(s)
        except Exception as e:
            raise commands.BadArgument("<duree>") from e

        if duree or not secondes:
            raise commands.BadArgument("<duree>")

        ts = datetime.datetime.now() + datetime.timedelta(seconds=secondes)
        action_id = None            # ID de l'action associée à la tâche (utile pour propagation à la suppression de l'action)
        try:
            quoi, id = commande.split(" ")
            if quoi in ["!open", "!close", "!remind"]:
                action_id = int(id)
        except ValueError:
            pass

        add_task(ctx.bot, ts, commande, action=action_id)

        await ctx.send(f"Commande {tools.code(commande)} planifiée pour le {tools.code(ts.strftime('%d/%m/%Y %H:%M:%S'))}")


    @commands.command()
    @tools.mjs_only
    async def cancel(self, ctx, *ids):
        """Annule une ou plusieurs tâche(s) planifiée(s) (COMMANDE MJ)

        [ids...] IDs des tâches à annuler, séparées par des espaces.
        Utiliser !taches pour voir la liste des IDs.
        """
        taches = [tache for id in ids if id.isdigit() and (tache := Taches.query.get(int(id)))]
        if taches:
            message = await ctx.send("Annuler les tâches :\n" + "\n".join([f" - {tools.code(tache.timestamp.strftime('%d/%m/%Y %H:%M:%S'))} > {tools.code(tache.commande)}" for tache in taches]))
            if await tools.yes_no(ctx.bot, message):
                for tache in taches:
                    cancel_task(ctx.bot, tache)

                await ctx.send("Tâche(s) annulée(s).")
            else:
                await ctx.send("Mission aborted.")
        else:
            await ctx.send(f"Aucune tâche trouvée.")



import json

from lgrez.blocs import env
import gspread
from oauth2client.service_account import ServiceAccountCredentials


def connect(key):
    """Charge les credentials GSheets et renvoie le classeur d'ID <key>"""
    # use creds to create a client to interact with the Google Drive API
    LGREZ_GCP_CREDENTIALS = env.load("LGREZ_GCP_CREDENTIALS")

    scope = ['https://spreadsheets.google.com/feeds']
    creds = ServiceAccountCredentials.from_json_keyfile_dict(json.loads(LGREZ_GCP_CREDENTIALS), scope)
    client = gspread.authorize(creds)

    # Open the workbook
    workbook = client.open_by_key(key)

    return workbook


def update(sheet, Modifs):
    """Met à jour la feuille <sheet> avec les modifications indiquées dans <Modifs>.

    <Modifs> doit être une liste de tuples (ligne (id), colonne (id), valeur)
    Les IDs sont indexés à partir de 0 (A1 en (0, 0, valeur)).
    """
    lm = max([l for (l, c, v) in Modifs])       # ligne max de la zone à modifier
    cm = max([c for (l, c, v) in Modifs])       # colonne max de la zone à modifier

    # Récupère toutes les valeurs sous forme de cellules gspread
    cells = sheet.range(1, 1, lm+1, cm+1)   # gspread indexe à partir de 1 (comme les gsheets)

    cells_to_update = []
    for (l, c, v) in Modifs:
        cell = [cell for cell in cells if cell.col == c+1 and cell.row == l+1][0]    # on récup l'objet Cell correspondant aux coords à modifier
        if isinstance(v, int) and v > 10**14:
            cell.value = str(v)
        elif v is None:
            cell.value = ""
        else:
            cell.value = v              # cells : ([<L1C1>, <L1C2>, ..., <L1Ccm>, <L2C1>, <L2C2>, ..., <LlmCcm>]
        cells_to_update.append(cell)

    sheet.update_cells(cells_to_update)
